#!/usr/bin/env python3
#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Script to generate Visual Studio Code configuration file.

This detects the Python include path and version, and the MSVC compiler path and
Windows SDK version. It then generates a "c_cpp_properties.json" file that can be
used by Visual Studio Code on the given system.
"""

import os
import sys

# Add Nuitka's inline copy of Jinja2 to sys.path
# Assuming this script is in 'misc/' and 'nuitka/' is in the root
repo_root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(0, repo_root)

# isort:start

import sysconfig

from nuitka.utils.FileOperations import getFileContents, putTextFileContents
from nuitka.utils.Jinja2 import getTemplateFromString
from nuitka.utils.Utils import isMacOS


def getPythonInfo():
    """Detects Python include path and version."""
    include_path = sysconfig.get_path("include")
    version_hex = hex((sys.version_info.major << 8) | (sys.version_info.minor << 4))

    return include_path, version_hex


def getMSVCInfo():
    """Detects MSVC compiler path and Windows SDK version."""
    # This is a simplified version of what Nuitka does,
    # leveraging the existing Utils if possible, or re-implementing if needed.
    # Nuitka's getMSVCRedistPath finds the redist folder, but we need the compiler.

    # Let's try to use vswhere directly as Nuitka does in Utils.py
    # but we need to find the compiler, not just the redist.
    # spell-checker: ignore vswhere

    # Return driven and many cases to deal with,
    # pylint: disable=too-many-branches,too-many-locals,too-many-return-statements
    vswhere_path = None

    for candidate in ("ProgramFiles(x86)", "ProgramFiles"):
        program_files_dir = os.getenv(candidate)
        if program_files_dir:
            candidate_path = os.path.join(
                program_files_dir,
                "Microsoft Visual Studio",
                "Installer",
                "vswhere.exe",
            )
            if os.path.exists(candidate_path):
                vswhere_path = candidate_path
                break

    if not vswhere_path:
        print("Error: vswhere.exe not found.")
        return None, None

    import subprocess

    # Find latest VS installation
    try:
        output = (
            subprocess.check_output(
                [
                    vswhere_path,
                    "-latest",
                    "-products",
                    "*",
                    "-requires",
                    "Microsoft.VisualStudio.Component.VC.Tools.x86.x64",
                    "-property",
                    "installationPath",
                ]
            )
            .decode()
            .strip()
        )
    except subprocess.CalledProcessError:
        print("Error: Failed to run vswhere.exe")
        return None, None

    if not output:
        print("Error: No suitable Visual Studio installation found.")
        return None, None

    vs_path = output

    # Find MSVC compiler
    vc_tools_version_file = os.path.join(
        vs_path, "VC", "Auxiliary", "Build", "Microsoft.VCToolsVersion.default.txt"
    )
    if os.path.exists(vc_tools_version_file):
        vc_version = getFileContents(vc_tools_version_file).strip()
    else:
        # Fallback: list directories in VC/Tools/MSVC and take the latest
        msvc_base = os.path.join(vs_path, "VC", "Tools", "MSVC")
        if os.path.exists(msvc_base):
            versions = sorted(os.listdir(msvc_base))
            if versions:
                vc_version = versions[-1]
            else:
                print("Error: No MSVC versions found.")
                return None, None
        else:
            print("Error: MSVC directory not found.")
            return None, None

    # Determine architecture, spell-checker: ignore Hostx64, Hostx86
    arch = "x64" if sys.maxsize > 2**32 else "x86"
    host_arch = (
        "Hostx64" if os.getenv("PROCESSOR_ARCHITECTURE") == "AMD64" else "Hostx86"
    )

    compiler_path = os.path.join(
        vs_path, "VC", "Tools", "MSVC", vc_version, "bin", host_arch, arch, "cl.exe"
    )

    if not os.path.exists(compiler_path):
        print(f"Error: Compiler not found at {compiler_path}")
        return None, None

    # Detect Windows SDK Version
    # This is a bit tricky without opening a developer command prompt.
    # We can look into the registry or check standard paths.
    # A common way is to look at the include paths in the registry or
    # check 'C:\Program Files (x86)\Windows Kits\10\Include'

    sdk_version = "10.0.19041.0"  # Default fallback

    kits_root = os.path.join(
        os.getenv("ProgramFiles(x86)", "C:\\Program Files (x86)"),
        "Windows Kits",
        "10",
        "Include",
    )
    if os.path.exists(kits_root):
        versions = [d for d in os.listdir(kits_root) if d.startswith("10.")]
        if versions:
            # Sort by version
            versions.sort(key=lambda s: [int(u) for u in s.split(".")])
            sdk_version = versions[-1]

    return compiler_path, sdk_version


def getPluginIncludePaths():
    """Scans for plugin directories containing C/C++ header files."""
    plugin_includes = []
    base_plugins_dir = os.path.join(repo_root, "nuitka", "plugins")

    if not os.path.exists(base_plugins_dir):
        return []

    for root, _, files in os.walk(base_plugins_dir):
        # We only care if the directory contains header files directly
        if any(f.endswith(".h") or f.endswith(".hpp") for f in files):
            rel_path = os.path.relpath(root, repo_root)
            # Ensure forward slashes for VSCode config
            path_str = "${workspaceFolder}/" + rel_path.replace("\\", "/")
            plugin_includes.append(path_str)

    return sorted(plugin_includes)


def getCompilerInfo():
    """Detects compiler path and IntelliSense mode based on OS."""
    if sys.platform == "win32":
        compiler_path, sdk_version = getMSVCInfo()
        return compiler_path, sdk_version, "msvc-x64"
    else:
        if isMacOS():
            mode = "macos-clang-x64"
            compiler_path = "/usr/bin/clang"
            if not os.path.exists(compiler_path) and os.path.exists("/usr/bin/gcc"):
                compiler_path = "/usr/bin/gcc"
        else:
            mode = "linux-gcc-x64"
            compiler_path = "/usr/bin/gcc"
            if not os.path.exists(compiler_path) and os.path.exists("/usr/bin/clang"):
                compiler_path = "/usr/bin/clang"

        return compiler_path, "", mode


def main():
    print("Generating .vscode/c_cpp_properties.json...")

    python_include, python_version = getPythonInfo()
    print(f"Python Include: {python_include}")
    print(f"Python Version: {python_version}")

    compiler_path, sdk_version, intelliSenseMode = getCompilerInfo()

    if not compiler_path:
        print("Failed to detect compiler.")
        return 1

    print(f"Compiler Path: {compiler_path}")
    if sdk_version:
        print(f"Windows SDK Version: {sdk_version}")
    print(f"IntelliSense Mode: {intelliSenseMode}")

    plugin_include_paths = getPluginIncludePaths()
    print(f"Found {len(plugin_include_paths)} plugins with headers.")

    template_path = os.path.join(repo_root, ".vscode", "c_cpp_properties.json.j2")
    template_str = getFileContents(template_path)

    # Render template
    # We use Nuitka's Jinja2 wrapper

    output = getTemplateFromString(template_str).render(
        python_include_path=python_include,
        python_version_hex=python_version,
        compiler_path=compiler_path,
        windows_sdk_version=sdk_version,
        intelliSenseMode=intelliSenseMode,
        plugin_include_paths=plugin_include_paths,
    )

    output_path = os.path.join(repo_root, ".vscode", "c_cpp_properties.json")
    putTextFileContents(output_path, output)

    print(f"Successfully generated {output_path}")
    return 0


if __name__ == "__main__":
    sys.exit(main())

#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Regression test for macOS PySide6 Qt WebEngine framework inclusion."""

# nuitka-project: --enable-plugin=pyside6
# nuitka-project: --mode=app
# nuitka-project: --macos-app-icon=none

# nuitka-skip-unless-expression: sys.platform == "darwin"
# nuitka-skip-unless-imports: PySide6.QtWebEngineCore

from __future__ import print_function

import os
import sys

from PySide6.QtCore import QTimer, QUrl
from PySide6.QtGui import QGuiApplication
from PySide6.QtWebEngineCore import QWebEnginePage

os.environ.setdefault("QTWEBENGINE_DISABLE_SANDBOX", "1")


# The macOS builder is headless, so avoid requiring a visible WebEngine widget.
app = QGuiApplication(sys.argv[:1] + ["-platform", "offscreen"])
page = QWebEnginePage()

html = """\
<!DOCTYPE html>
<html>
<head>
  <title>qtwe3857</title>
  <script>
    function getAnswer() {
      return document.getElementById("answer").textContent;
    }
  </script>
</head>
<body>
  <h1 id="answer">3857</h1>
</body>
</html>
"""

state = {"error": None, "timed_out": False}


def onFailure(message):
    state["error"] = message
    print("failure=%s" % message)
    sys.stdout.flush()
    app.quit()


def onJavaScriptDone(result):
    print("js-result=%s" % result)

    if result != "3857":
        onFailure("Unexpected JavaScript result '%s'" % result)
        return

    sys.stdout.flush()
    app.quit()


def onLoadFinished(ok):
    print("loaded=%s" % ok)

    if not ok:
        onFailure("WebEngine failed to load inline HTML")
        return

    title = page.title()
    print("title=%s" % title)

    if title != "qtwe3857":
        onFailure("Unexpected page title '%s'" % title)
        return

    page.runJavaScript("getAnswer()", onJavaScriptDone)


def onTimeout():
    state["timed_out"] = True
    print("timeout")
    sys.stdout.flush()
    app.quit()


page.loadFinished.connect(onLoadFinished)
QTimer.singleShot(15000, onTimeout)
page.setHtml(html, QUrl("file:///qtwe3857.html"))
app.exec()

if state["error"] is not None:
    raise AssertionError(state["error"])

if state["timed_out"]:
    raise AssertionError("WebEngine timed out")

#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#!/usr/bin/env python
#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Main program for codespell checker tool."""

import os
import re
import sys

from nuitka.format.FileFormatting import cleanupWindowsNewlines
from nuitka.options.CommandLineOptionsTools import makeOptionsParser
from nuitka.tools.Basics import goHome
from nuitka.tools.quality.ScanSources import scanTargets
from nuitka.Tracing import my_print, tools_logger
from nuitka.utils.Execution import (
    callProcessChunked,
    check_output,
    getExecutablePath,
)
from nuitka.utils.FileOperations import (
    areSamePaths,
    getFileContents,
    putTextFileContents,
    resolveShellPatternToFilenames,
)

replacements = [
    ("organizational", "organisational"),
    ("developer manual", "Developer Manual"),
    ("user manual", "User Manual"),
]


def _isSpellcheckerIgnoringFile(filename):
    return "spell-checker: disable" in getFileContents(filename)


def _isGeneratedFile(contents):
    for line in contents.splitlines()[:20]:
        if "WARNING, this code is GENERATED" in line:
            return True
    return False


def _checkIgnoreWords(contents):
    ignored_words = []
    lines = contents.splitlines()
    clean_lines = []

    for line in lines:
        match = re.search(r"spell-checker:\s*ignore\s+(.*)", line)
        if match:
            # Add words found to the list of ignored words, we need to check matches
            # for them.
            words = match.group(1).replace(",", " ").split()
            ignored_words.extend(words)

            # verify if the word is present in the file, checking the line too, but
            # ignoring the command itself.
            line = line[: match.start()] + line[match.end() :]

        clean_lines.append(line)

    if not ignored_words:
        return []

    clean_content = "\n".join(clean_lines).lower()

    unused = []
    for word in ignored_words:
        # Check if the word is present in the file, as a substring (e.g. CamelCase).
        if word.lower() not in clean_content:
            unused.append(word)

    return unused


def runCodespell(filenames, verbose, write):
    if verbose:
        my_print("Consider", " ".join(filenames))

    if os.name == "nt":
        extra_path = os.path.join(sys.prefix, "Scripts")
    else:
        extra_path = None

    codespell_binary = getExecutablePath("codespell", extra_dir=extra_path)

    codespell_version = check_output([codespell_binary, "--version"])

    if str is not bytes:
        codespell_version = codespell_version.decode("utf8").strip()

    my_print("Using codespell version:", codespell_version)

    command = [
        codespell_binary,
        "-f",
        "-I",
        os.path.join(
            os.path.dirname(__file__),
            "..",
            "..",
            "..",
            "..",
            "misc/codespell-ignore.txt",
        ),
    ]

    if write:
        command.append("-w")

    result = callProcessChunked(
        command, filenames, logger=tools_logger if verbose else None
    )

    # Check for unnecessary ignores, but only if the file is cleanly passing
    # codespell check itself.
    found_superfluous_ignores = False

    if result == 0:
        for filename in filenames:
            if areSamePaths(__file__, filename):
                continue

            contents = getFileContents(filename)

            if not _isGeneratedFile(contents):
                unused_ignores = _checkIgnoreWords(contents)

                if unused_ignores:
                    my_print(
                        "%s: Error, unused ignore words: %s"
                        % (filename, ",".join(unused_ignores))
                    )
                    found_superfluous_ignores = True

            old_contents = contents
            for word, replacement in replacements:
                contents = contents.replace(word, replacement)
                contents = contents.replace(word.title(), replacement.title())

            if old_contents != contents:
                putTextFileContents(filename, contents)
                cleanupWindowsNewlines(filename, filename)

    return result == 0 and not found_superfluous_ignores


def main():
    parser = makeOptionsParser(usage="%prog [options]", epilog=None)

    parser.add_option(
        "--verbose",
        action="store_true",
        dest="verbose",
        default=False,
        help="""Show what it is doing. Default is %default.""",
    )

    parser.add_option(
        "--write",
        "-w",
        action="store_true",
        dest="write",
        default=False,
        help="""Write changes to the files. Default is %default.""",
    )

    options, positional_args = parser.parse_args()

    if not positional_args:
        positional_args = [
            "bin",
            "nuitka",
            "rpm",
            "misc",
            "tests/*/run_all.py",
            "*.rst",
        ]
        goHome()

    my_print("Working on:", positional_args)

    positional_args = sum(
        (
            resolveShellPatternToFilenames(positional_arg)
            for positional_arg in positional_args
        ),
        [],
    )

    filenames = list(
        scanTargets(
            positional_args,
            suffixes=(".py", ".scons", ".rst", ".txt", ".j2", ".md", ".c", ".h"),
            ignore_list=("get-pip-2.6.py",),
        )
    )

    filenames = [
        filename for filename in filenames if not _isSpellcheckerIgnoringFile(filename)
    ]
    if not filenames:
        sys.exit("No files found.")

    result = runCodespell(
        filenames=filenames, verbose=options.verbose, write=options.write
    )

    if result:
        my_print("OK.")
    else:
        my_print("FAILED.")
        tools_logger.sysexit(
            "\nError, please correct the spelling problems found or extend 'misc/codespell-ignore.txt' if applicable."
        )


if __name__ == "__main__":
    main()

#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Textual standalone test with lazy-loaded widgets."""

# nuitka-skip-unless-imports: textual

# nuitka-project: --mode=standalone

from __future__ import print_function

from textual.app import App
from textual.widgets import Footer, Header, Static


class TextualUsingApp(App):
    def compose(self):
        yield Header()
        yield Static("Textual body", id="body")
        yield Footer()

    def on_ready(self):
        header = self.query_one(Header)
        body = self.query_one(Static)
        footer = self.query_one(Footer)

        print("Header module:", Header.__module__)
        print("Static module:", Static.__module__)
        print("Footer module:", Footer.__module__)
        print(
            "Mounted widgets:",
            ",".join(widget.__class__.__name__ for widget in (header, body, footer)),
        )
        print("Body id:", body.id)

        self.exit()


if __name__ == "__main__":
    TextualUsingApp().run(headless=True)
    print("OK.")

#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""This is the main actions of Nuitka.

This can do all the steps to translate one module to a target language using
the Python C/API, to compile it to either an executable or an extension
module, potentially with bytecode included and used library copied into
a distribution folder.

"""

import os
import sys

from nuitka.build.AdaptPythonHeaderFiles import createAdaptedPythonHeaderFiles
from nuitka.build.DataComposerInterface import runDataComposer
from nuitka.build.SconsInterface import provideStaticSourceFilesBackend
from nuitka.build.SconsUtils import (
    getSconsCompilerUsed,
    getSconsReportValue,
    readSconsErrorReport,
    readSconsObjectSizes,
    readSconsReport,
    readSconsResourceUsageReports,
)
from nuitka.code_generation.CodeGeneration import (
    generateHelpersCode,
    generateModuleCode,
)
from nuitka.code_generation.ConstantCodes import (
    addDistributionMetadataValue,
    getDistributionMetadataValues,
)
from nuitka.freezer.IncludedDataFiles import (
    addIncludedDataFilesFromFileOptions,
    addIncludedDataFilesFromFlavor,
    addIncludedDataFilesFromPackageOptions,
    addIncludedDataFilesFromPlugins,
    checkProjectExpectedDataFiles,
    copyDataFiles,
)
from nuitka.freezer.IncludedEntryPoints import (
    addExtensionModuleEntryPoint,
    addIncludedEntryPoints,
    addMainEntryPoint,
    getStandaloneEntryPoints,
)
from nuitka.freezer.MacOSApp import addIncludedDataFilesFromMacOSAppOptions
from nuitka.freezer.MacOSDmg import createDmgFile
from nuitka.importing.Importing import (
    getRecompileDecisionReason,
    locateModule,
    setupImportingFromOptions,
)
from nuitka.importing.Recursion import (
    scanIncludedPackage,
    scanPluginFilenamePattern,
    scanPluginPath,
    scanPluginSinglePath,
)
from nuitka.optimizations.ValueTraces import setupValueTraceFromOptions
from nuitka.options.Options import (
    assumeYesForDownloads,
    getDebuggerName,
    getFileReferenceMode,
    getForcedStderrPath,
    getForcedStdoutPath,
    getMainArgs,
    getMainEntryPointFilenames,
    getMustIncludeModules,
    getMustIncludePackages,
    getOutputDir,
    getPgoArgs,
    getPositionalArgs,
    getPythonPgoInput,
    getShallFollowExtra,
    getShallFollowExtraFilePatterns,
    getShallFollowModules,
    getShallIncludeDistributionMetadata,
    getXMLDumpOutputFilename,
    hasPythonFlagIsolated,
    hasPythonFlagNoAnnotations,
    hasPythonFlagNoAsserts,
    hasPythonFlagNoBytecodeRuntimeCache,
    hasPythonFlagNoCurrentDirectoryInPath,
    hasPythonFlagNoDocStrings,
    hasPythonFlagNoRandomization,
    hasPythonFlagNoSite,
    hasPythonFlagNoWarnings,
    hasPythonFlagTraceImports,
    hasPythonFlagUnbuffered,
    isCompileTimeProfile,
    isCPgoMode,
    isExperimental,
    isLowMemory,
    isMultidistMode,
    isOnefileMode,
    isRemoveBuildDir,
    isRuntimeProfile,
    isShowInclusion,
    isShowMemory,
    isShowProgress,
    isStandaloneMode,
    shallAskForWindowsAdminRights,
    shallCreateDmgFile,
    shallCreatePythonPgoInput,
    shallCreateScriptFileForExecution,
    shallExecuteImmediately,
    shallMakeModule,
    shallMakePackage,
    shallNotDoExecCCompilerCall,
    shallOnlyExecCCompilerCall,
    shallRunInDebugger,
    shallTraceExecution,
    shallTreatUninstalledPython,
    shallUsePythonDebug,
    shallUseStaticLibPython,
)
from nuitka.plugins.Hooks import (
    considerExtraDlls,
    getBuildDefinitions,
    onBeforeCodeParsing,
    onCompilationStartChecks,
    onFinalResult,
    onGeneratedSourceCode,
    onModuleCompleteSet,
    onModuleInitialSet,
    onStandaloneDistributionFinished,
    writeExtraCodeFiles,
)
from nuitka.plugins.PluginsUsage import printPluginUsageStats
from nuitka.PostProcessing import executePostProcessing
from nuitka.Progress import (
    closeProgressBar,
    reportProgressBar,
    setupProgressBar,
)
from nuitka.PythonFlavors import (
    getPythonFlavorName,
    isApplePython,
    isArchPackagePython,
    isDebianPackagePython,
    isFedoraPackagePython,
    isMonolithPy,
    isPyenvPython,
)
from nuitka.PythonVersions import (
    getModuleLinkerLibs,
    getPythonABI,
    getSupportedPythonVersions,
    isPythonWithGil,
    python_version,
    python_version_str,
)
from nuitka.Serialization import ConstantAccessor
from nuitka.Tracing import (
    code_generation_logger,
    doNotBreakSpaces,
    general,
    inclusion_logger,
    pgo_logger,
)
from nuitka.tree import SyntaxErrors
from nuitka.tree.ReformulationMultidist import createMultidistMainSourceCode
from nuitka.utils.Distributions import getDistribution, getDistributionName
from nuitka.utils.Execution import (
    callProcess,
    withEnvironmentVarOverridden,
    wrapCommandForDebuggerForExec,
)
from nuitka.utils.FileOperations import (
    changeFilenameExtension,
    deleteFile,
    getExternalUsePath,
    getNormalizedPathJoin,
    getReportPath,
    isFilesystemEncodable,
    openTextFile,
    removeDirectory,
)
from nuitka.utils.Importing import getPackageDirFilename
from nuitka.utils.InstanceCounters import printInstanceCounterStats
from nuitka.utils.MemoryUsage import reportMemoryUsage, showMemoryTrace
from nuitka.utils.ModuleNames import ModuleName
from nuitka.utils.ReExecute import callExecProcess, reExecuteNuitka
from nuitka.utils.StaticLibraries import getSystemStaticLibPythonPath
from nuitka.utils.Timing import withProfiling
from nuitka.utils.Utils import getArchitecture, isMacOS, isWin32Windows
from nuitka.Version import getCommercialVersion, getNuitkaVersion

from . import ModuleRegistry, OutputDirectories
from .build.SconsInterface import (
    asBoolStr,
    cleanSconsDirectory,
    getCommonSconsOptions,
    runScons,
)
from .code_generation import LoaderCodes, Reports
from .finalizations import Finalization
from .freezer.Onefile import getCompressorPython, packDistFolderToOnefile
from .freezer.Standalone import (
    checkFreezingModuleSet,
    copyDllsUsed,
    detectUsedDLLs,
    signDistributionMacOS,
)
from .optimizations.Optimization import optimizeModules
from .pgo.PGO import readPGOInputFile
from .reports.Reports import writeCompilationReports
from .States import states
from .tree.Building import buildMainModuleTree
from .tree.SourceHandling import writeSourceCode
from .TreeXML import dumpTreeXMLToFile


def _deleteResultFile(filename):
    deleteFile(
        path=filename,
        must_exist=False,
    )
    if isWin32Windows():
        deleteFile(
            path=changeFilenameExtension(filename, ".pdb"),
            must_exist=False,
        )


def _createMainModule():
    """Create a node tree.

    Turn that source code into a node tree structure. If following into
    imported modules is allowed, more trees will be available during
    optimization, or even immediately through forcefully included
    directory paths.

    """
    # Many cases and details to deal with, pylint: disable=too-many-branches

    onBeforeCodeParsing()

    # First, build the raw node tree from the source code.
    if isMultidistMode():
        assert not shallMakeModule()

        main_module = buildMainModuleTree(
            source_code=createMultidistMainSourceCode(),
        )
    else:
        main_module = buildMainModuleTree(
            source_code=None,
        )

    OutputDirectories.setMainModule(main_module)

    for distribution_name in getShallIncludeDistributionMetadata():
        distribution = getDistribution(distribution_name)

        if distribution is None:
            return general.sysexit(
                "Error, could not find distribution '%s' for which metadata was asked to be included."
                % distribution_name
            )

        real_distribution_name = getDistributionName(distribution)

        if real_distribution_name != distribution_name:
            general.warning("""\
Warning, the distribution specified as '--include-distribution-metadata=%s' is really named '%s', \
use the correct name instead.""" % (distribution_name, real_distribution_name))

        addDistributionMetadataValue(
            distribution_name=real_distribution_name,
            distribution=distribution,
            reason="user requested",
        )

    # First remove old object files and old generated files, old binary or
    # module, and standalone mode program directory if any, they can only do
    # harm.
    source_dir = OutputDirectories.getSourceDirectoryPath(onefile=False, create=True)

    if not shallOnlyExecCCompilerCall():
        cleanSconsDirectory(source_dir)

        # Prepare the ".dist" directory, throwing away what was there before.
        if isStandaloneMode():
            OutputDirectories.initStandaloneDirectory(logger=general)

    # Delete result file, to avoid confusion with previous build and to
    # avoid locking issues after the build.
    _deleteResultFile(OutputDirectories.getResultFullpath(onefile=False, real=True))

    if isOnefileMode():
        _deleteResultFile(OutputDirectories.getResultFullpath(onefile=True, real=True))

        # Also make sure we inform the user in case the compression is not possible.
        getCompressorPython()

    # Second, do it for the directories given.
    for plugin_filename in getShallFollowExtra():
        scanPluginPath(plugin_filename=plugin_filename, module_package=None)

    for pattern in getShallFollowExtraFilePatterns():
        scanPluginFilenamePattern(pattern=pattern)

    # For packages, include the full suite.
    if shallMakePackage():
        scanIncludedPackage(main_module.getFullName())

    for package_name in getMustIncludePackages():
        scanIncludedPackage(package_name)

    for module_name in getMustIncludeModules():
        module_name, module_filename, module_kind, finding = locateModule(
            module_name=ModuleName(module_name),
            parent_package=None,
            level=0,
        )

        if finding != "absolute":
            return inclusion_logger.sysexit(
                "Error, failed to locate module '%s' that you asked to include."
                % module_name.asString()
            )

        if module_kind == "built-in":
            # TODO:
            inclusion_logger.warning(
                "Note, module '%s' that you asked to include is built-in."
                % module_name.asString()
            )
        else:
            scanPluginSinglePath(
                plugin_filename=module_filename,
                module_package=module_name.getPackageName(),
                package_only=True,
            )

    # Allow plugins to add more modules based on the initial set being complete.
    onModuleInitialSet()

    # Then optimize the tree and potentially recursed modules.
    # TODO: The passed filename is really something that should come from
    # a command line option, it's a filename for the graph, which might not
    # need a default at all.
    optimizeModules(main_module.getOutputFilename())

    # Freezer may have concerns for some modules.
    if isStandaloneMode():
        checkFreezingModuleSet()

    # Check if distribution meta data is included, that cannot be used.
    for distribution_name, meta_data_value in getDistributionMetadataValues():
        if not ModuleRegistry.hasDoneModule(meta_data_value.module_name):
            return inclusion_logger.sysexit(
                "Error, including metadata for distribution '%s' without including related package '%s'."
                % (distribution_name, meta_data_value.module_name)
            )

    # Complain about undecided recompilations of extension modules
    for module in ModuleRegistry.getDoneModules():
        recompile_decision = getRecompileDecisionReason(module.getFullName())

        # TODO: Make that a shared constant value
        if (
            recompile_decision[0] is None
            and recompile_decision[1] == "default behavior"
        ):
            inclusion_logger.info("""\
Should decide '--prefer-source-code' vs. '--no-prefer-source-code', using \
existing '%s' extension module by default, but source code is available and \
may work too.""" % (module.getFullName()))

    # Allow plugins to comment on final module set.
    onModuleCompleteSet()

    if isExperimental("check_xml_persistence"):
        for module in ModuleRegistry.getRootModules():
            if module.isMainModule():
                return module

        assert False
    else:
        # Main module might change behind our back, look it up again.
        return main_module


def dumpTreeXML():
    filename = getXMLDumpOutputFilename()

    if filename is not None:
        with openTextFile(filename, "wb") as output_file:
            # XML output only.
            for module in ModuleRegistry.getDoneModules():
                dumpTreeXMLToFile(tree=module.asXml(), output_file=output_file)

        general.info("XML dump of node state written to file '%s'." % filename)


def pickSourceFilenames(source_dir, modules):
    """Pick the names for the C files of each module.

    Args:
        source_dir - the directory to put the module sources will be put into
        modules    - all the modules to build.

    Returns:
        Dictionary mapping modules to filenames in source_dir.

    Notes:
        These filenames can collide, due to e.g. mixed case usage, or there
        being duplicate copies, e.g. a package named the same as the main
        binary.

        Conflicts are resolved by appending @<number> with a count in the
        list of sorted modules. We try to be reproducible here, so we get
        still good caching for external tools.
    """

    collision_filenames = set()

    def _getModuleFilenames(module):
        nice_filename = getNormalizedPathJoin(
            source_dir, "module." + module.getFullName()
        )

        # Note: Could detect if the file system is cases sensitive in source_dir
        # or not, but that's probably not worth the effort. False positives do
        # no harm at all. We cannot use normcase, as macOS is not using one that
        # will tell us the truth.
        collision_filename = getNormalizedPathJoin(
            source_dir, "module." + module.getFullName().asString().lower()
        )

        # When the filename becomes to long to add ".const", we use a hash name
        # instead.
        hash_filename = getNormalizedPathJoin(
            source_dir,
            "module.hashed_" + module.getFullName().asLegalFilename(name_limit=1),
        )

        return nice_filename, collision_filename, hash_filename

    colliding_filenames = set()

    # First pass, check for collisions.
    for module in modules:
        if module.isPythonExtensionModule():
            continue

        _nice_filename, collision_filename, _hash_filename = _getModuleFilenames(module)

        if collision_filename in colliding_filenames:
            collision_filenames.add(collision_filename)

        colliding_filenames.add(collision_filename)

    # Our output.
    module_filenames = {}

    # Second pass, handle collisions and encoding issues for filenames.
    for module in modules:
        if module.isPythonExtensionModule():
            continue

        nice_filename, collision_filename, hash_filename = _getModuleFilenames(module)

        base_filename = os.path.basename(nice_filename)

        # Allow for longer suffixes that .c, we use .const and might use others
        # as well in the C compiler and make sure we use only file system
        # encodable names.
        if (
            collision_filename in collision_filenames
            or len(base_filename) > 240
            or not isFilesystemEncodable(base_filename)
        ):
            nice_filename = hash_filename

        module_filenames[module] = nice_filename + ".c"

    return module_filenames


def makeSourceDirectory():
    """Get the full list of modules imported, create code for all of them."""
    # We deal with a lot of details here, but rather one by one, and split makes
    # no sense, pylint: disable=too-many-branches

    # assert main_module in ModuleRegistry.getDoneModules()

    # Lets check if the asked modules are actually present, and warn the
    # user if one of those was not found.
    for any_case_module in getShallFollowModules():
        if "*" in any_case_module or "{" in any_case_module:
            continue

        if not ModuleRegistry.hasDoneModule(
            any_case_module
        ) and not ModuleRegistry.hasRootModule(any_case_module):
            general.warning(
                "Did not follow import to unused '%s', consider include "
                % any_case_module
            )

    # Prepare code generation, i.e. execute finalization for it.
    for current_module in ModuleRegistry.getDoneModules():
        if current_module.isCompiledPythonModule():
            Finalization.prepareCodeGeneration(current_module)

    # Do some reporting and determine compiled module to work on
    compiled_modules = []

    for current_module in ModuleRegistry.getDoneModules():
        if current_module.isCompiledPythonModule():
            compiled_modules.append(current_module)

            if isShowInclusion():
                inclusion_logger.info(
                    "Included compiled module '%s'." % current_module.getFullName()
                )
        elif current_module.isPythonExtensionModule():
            if isShowInclusion():
                inclusion_logger.info(
                    "Included extension module '%s'." % current_module.getFullName()
                )
        elif current_module.isUncompiledPythonModule():
            if isShowInclusion():
                inclusion_logger.info(
                    "Included uncompiled module '%s'." % current_module.getFullName()
                )
        else:
            assert False, current_module

    # Pick filenames.
    source_dir = OutputDirectories.getSourceDirectoryPath(onefile=False, create=False)

    module_filenames = pickSourceFilenames(
        source_dir=source_dir, modules=compiled_modules
    )

    setupProgressBar(
        stage="C Source Generation",
        unit="module",
        total=len(compiled_modules),
    )

    # Generate code for compiled modules, this can be slow, so do it separately
    # with a progress bar.
    for current_module in compiled_modules:
        module_name = current_module.getFullName()
        c_filename = module_filenames[current_module]

        reportProgressBar(
            item=module_name,
        )

        source_code = generateModuleCode(
            module=current_module,
            data_filename=changeFilenameExtension(
                os.path.basename(c_filename), ".const"
            ),
        )

        writeSourceCode(
            filename=c_filename,
            source_code=source_code,
            logger=code_generation_logger,
            assume_yes_for_downloads=assumeYesForDownloads(),
        )

    closeProgressBar()

    (
        helper_decl_code,
        helper_impl_code,
        constants_header_code,
        constants_body_code,
    ) = generateHelpersCode()

    writeSourceCode(
        filename=getNormalizedPathJoin(source_dir, "__helpers.h"),
        source_code=helper_decl_code,
        logger=code_generation_logger,
        assume_yes_for_downloads=assumeYesForDownloads(),
    )

    writeSourceCode(
        filename=getNormalizedPathJoin(source_dir, "__helpers.c"),
        source_code=helper_impl_code,
        logger=code_generation_logger,
        assume_yes_for_downloads=assumeYesForDownloads(),
    )

    writeSourceCode(
        filename=getNormalizedPathJoin(source_dir, "__constants.h"),
        source_code=constants_header_code,
        logger=code_generation_logger,
        assume_yes_for_downloads=assumeYesForDownloads(),
    )

    writeSourceCode(
        filename=getNormalizedPathJoin(source_dir, "__constants.c"),
        source_code=constants_body_code,
        logger=code_generation_logger,
        assume_yes_for_downloads=assumeYesForDownloads(),
    )

    return module_filenames


def _runPgoBinary():
    pgo_executable = OutputDirectories.getPgoRunExecutable()

    if not os.path.isfile(pgo_executable):
        return general.sysexit(
            "Error, failed to produce PGO binary '%s'" % pgo_executable
        )

    return callProcess(
        [getExternalUsePath(pgo_executable)] + getPgoArgs(),
        shell=False,
    )


def _deleteMsvcPGOFiles(pgo_mode):
    assert getSconsCompilerUsed(
        OutputDirectories.getSourceDirectoryPath(onefile=False, create=False)
    ) in ("MSVC", "ClangCL")

    msvc_pgc_filename = OutputDirectories.getResultBasePath(onefile=False) + "!1.pgc"
    deleteFile(msvc_pgc_filename, must_exist=False)

    if pgo_mode == "use":
        msvc_pgd_filename = OutputDirectories.getResultBasePath(onefile=False) + ".pgd"
        deleteFile(msvc_pgd_filename, must_exist=False)

    return msvc_pgc_filename


def _runCPgoBinary():
    # Note: For exit codes, we do not insist on anything yet, maybe we could point it out
    # or ask people to make scripts that buffer these kinds of errors, and take an error
    # instead as a serious failure.

    general.info(
        "Running created binary to produce C level PGO information:", style="blue"
    )

    if getSconsCompilerUsed(
        OutputDirectories.getSourceDirectoryPath(onefile=False, create=False)
    ) in ("MSVC", "ClangCL"):
        msvc_pgc_filename = _deleteMsvcPGOFiles(pgo_mode="generate")

        with withEnvironmentVarOverridden(
            "PATH",
            getSconsReportValue(
                source_dir=OutputDirectories.getSourceDirectoryPath(
                    onefile=False, create=False
                ),
                key="PATH",
            ),
        ):
            exit_code_pgo = _runPgoBinary()

        pgo_data_collected = os.path.exists(msvc_pgc_filename)
    else:
        exit_code_pgo = _runPgoBinary()

        # gcc file suffix, spell-checker: ignore gcda
        gcc_constants_pgo_filename = getNormalizedPathJoin(
            OutputDirectories.getSourceDirectoryPath(onefile=False, create=False),
            "__constants.gcda",
        )

        pgo_data_collected = os.path.exists(gcc_constants_pgo_filename)

    if exit_code_pgo != 0:
        pgo_logger.warning("""\
Error, the C PGO compiled program error exited. Make sure it works \
fully before using '--pgo-c' option.""")

    if not pgo_data_collected:
        return pgo_logger.sysexit("""\
Error, no C PGO compiled program did not produce expected information, \
did the created binary run at all?""")

    pgo_logger.info("Successfully collected C level PGO information.", style="blue")


def _runPythonPgoBinary():
    # Note: For exit codes, we do not insist on anything yet, maybe we could point it out
    # or ask people to make scripts that buffer these kinds of errors, and take an error
    # instead as a serious failure.

    pgo_filename = OutputDirectories.getPgoRunInputFilename()

    with withEnvironmentVarOverridden("NUITKA_PGO_OUTPUT", pgo_filename):
        exit_code = _runPgoBinary()

    if not os.path.exists(pgo_filename):
        return general.sysexit("""\
Error, no Python PGO information produced, did the created binary
run (exit code %d) as expected?""" % exit_code)

    return pgo_filename


def runSconsBackend():
    # Scons gets transported many details, that we express as variables, and
    # have checks for them, leading to many branches and statements,
    # pylint: disable=too-many-branches,too-many-statements
    scons_options, env_values = getCommonSconsOptions()

    scons_options["source_dir"] = OutputDirectories.getSourceDirectoryPath(
        onefile=False, create=False
    )

    # We might need to adapt the Python header files for some setups to work correctly.
    adapted_dir = createAdaptedPythonHeaderFiles(scons_options["source_dir"])
    if adapted_dir:
        scons_options["adapted_python_header_files_dir"] = adapted_dir

    scons_options["monolithpy"] = asBoolStr(isMonolithPy())
    scons_options["debug_mode"] = asBoolStr(states.is_debug)
    scons_options["debugger_mode"] = asBoolStr(shallRunInDebugger())
    scons_options["python_debug"] = asBoolStr(shallUsePythonDebug())
    scons_options["full_compat"] = asBoolStr(states.is_full_compat)
    scons_options["trace_mode"] = asBoolStr(shallTraceExecution())
    scons_options["file_reference_mode"] = getFileReferenceMode()
    scons_options["compiled_module_count"] = "%d" % len(
        ModuleRegistry.getCompiledModules()
    )

    if isLowMemory():
        scons_options["low_memory"] = asBoolStr(True)

    scons_options["result_exe"] = OutputDirectories.getResultFullpath(
        onefile=False, real=False
    )

    if not shallMakeModule():
        main_module = ModuleRegistry.getRootTopModule()
        assert main_module.isMainModule()

        main_module_name = main_module.getFullName()
        if main_module_name != "__main__":
            scons_options["main_module_name"] = main_module_name

    if shallUseStaticLibPython():
        scons_options["static_libpython"] = getSystemStaticLibPythonPath(
            python_debug=shallUsePythonDebug()
        )

    if isDebianPackagePython():
        scons_options["debian_python"] = asBoolStr(True)
    if isFedoraPackagePython():
        scons_options["fedora_python"] = asBoolStr(True)
    if isArchPackagePython():
        scons_options["arch_python"] = asBoolStr(True)
    if isApplePython():
        scons_options["apple_python"] = asBoolStr(True)
    if isPyenvPython():
        scons_options["pyenv_python"] = asBoolStr(True)

    if getForcedStdoutPath():
        scons_options["forced_stdout_path"] = getForcedStdoutPath()

    if getForcedStderrPath():
        scons_options["forced_stderr_path"] = getForcedStderrPath()

    if isRuntimeProfile():
        scons_options["profile_mode"] = asBoolStr(True)

    if shallTreatUninstalledPython():
        scons_options["uninstalled_python"] = asBoolStr(True)

    if ModuleRegistry.getUncompiledTechnicalModules():
        scons_options["frozen_modules"] = str(
            len(ModuleRegistry.getUncompiledTechnicalModules())
        )

    if hasPythonFlagNoWarnings():
        scons_options["no_python_warnings"] = asBoolStr(True)

    if hasPythonFlagNoAsserts():
        scons_options["python_sysflag_optimize"] = str(
            2 if hasPythonFlagNoDocStrings() else 1
        )

        scons_options["python_flag_no_asserts"] = asBoolStr(True)

    if hasPythonFlagNoDocStrings():
        scons_options["python_flag_no_docstrings"] = asBoolStr(True)

    if hasPythonFlagNoAnnotations():
        scons_options["python_flag_no_annotations"] = asBoolStr(True)

    if python_version < 0x300 and sys.flags.py3k_warning:
        scons_options["python_sysflag_py3k_warning"] = asBoolStr(True)

    if python_version < 0x300 and (
        sys.flags.division_warning or sys.flags.py3k_warning
    ):
        scons_options["python_sysflag_division_warning"] = asBoolStr(True)

    if sys.flags.bytes_warning:
        scons_options["python_sysflag_bytes_warning"] = asBoolStr(True)

    # TODO: Actually all of sys flags is not for module mode and we should
    # achieve it to be an error to even attempt to use it.
    if not shallMakeModule() and int(
        os.getenv("NUITKA_NOSITE_FLAG", hasPythonFlagNoSite())
    ):
        scons_options["python_sysflag_no_site"] = asBoolStr(True)

    if hasPythonFlagTraceImports():
        scons_options["python_sysflag_verbose"] = asBoolStr(True)

    if hasPythonFlagNoRandomization():
        scons_options["python_sysflag_no_randomization"] = asBoolStr(True)

    if python_version < 0x300 and sys.flags.unicode:
        scons_options["python_sysflag_unicode"] = asBoolStr(True)

    if python_version >= 0x370 and sys.flags.utf8_mode:
        scons_options["python_sysflag_utf8"] = asBoolStr(True)

    if hasPythonFlagNoBytecodeRuntimeCache():
        scons_options["python_sysflag_dontwritebytecode"] = asBoolStr(True)

    if hasPythonFlagNoCurrentDirectoryInPath():
        scons_options["python_sysflag_safe_path"] = asBoolStr(True)

    if hasPythonFlagUnbuffered():
        scons_options["python_sysflag_unbuffered"] = asBoolStr(True)

    if hasPythonFlagIsolated():
        scons_options["python_sysflag_isolated"] = asBoolStr(True)

    abiflags = getPythonABI(python_debug=shallUsePythonDebug())
    if abiflags:
        scons_options["abiflags"] = abiflags

    link_module_libs = getModuleLinkerLibs()
    if link_module_libs:
        scons_options["link_module_libs"] = ",".join(link_module_libs)

    # Allow plugins to build definitions.
    env_values.update(getBuildDefinitions())

    if shallCreatePythonPgoInput():
        scons_options["pgo_mode"] = "python"

        result = runScons(
            scons_options=scons_options,
            env_values=env_values,
            scons_filename="Backend.scons",
        )

        if not result:
            return result, scons_options

        # Need to make it usable before executing it.
        executePostProcessing(scons_options["result_exe"])
        _runPythonPgoBinary()

        return True, scons_options

        # Need to restart compilation from scratch here.
    if isCPgoMode():
        # For C level PGO, we have a 2 pass system. TODO: Make it more global for onefile
        # and standalone mode proper support, which might need data files to be
        # there, which currently are not yet there, so it won't run.
        if isCPgoMode():
            scons_options["pgo_mode"] = "generate"

            result = runScons(
                scons_options=scons_options,
                env_values=env_values,
                scons_filename="Backend.scons",
            )

            if not result:
                return result, scons_options

            # Need to make it usable before executing it.
            executePostProcessing(scons_options["result_exe"])
            _runCPgoBinary()
            scons_options["pgo_mode"] = "use"

    result = (
        runScons(
            scons_options=scons_options,
            env_values=env_values,
            scons_filename="Backend.scons",
        ),
        scons_options,
    )

    # Delete PGO files if asked to do that.
    if scons_options.get("pgo_mode") == "use" and (
        getSconsCompilerUsed(
            OutputDirectories.getSourceDirectoryPath(onefile=False, create=False)
        )
        in ("MSVC", "ClangCL")
    ):
        _deleteMsvcPGOFiles(pgo_mode="use")

    return result


def callExecPython(args, add_path, uac):
    if add_path:
        if "PYTHONPATH" in os.environ:
            os.environ["PYTHONPATH"] += ":" + getOutputDir()
        else:
            os.environ["PYTHONPATH"] = getOutputDir()

    # Add the main arguments, previous separated.
    args += getPositionalArgs()[1:] + getMainArgs()

    callExecProcess(args, shell=uac)


def _executeMain(binary_filename):
    # Wrap in debugger, unless the CMD file contains that call already.
    if shallRunInDebugger() and not shallCreateScriptFileForExecution():
        args = wrapCommandForDebuggerForExec(
            command=(binary_filename,),
            debugger=getDebuggerName(),
        )
    else:
        args = (binary_filename, binary_filename)

    callExecPython(
        args=args,
        add_path=False,
        uac=isWin32Windows() and shallAskForWindowsAdminRights(),
    )


def _executeModule(tree):
    """Execute the extension module just created."""

    if python_version < 0x300:
        python_command_template = """\
import os, imp;\
assert os.path.normcase(os.path.abspath(os.path.normpath(\
imp.find_module('%(module_name)s')[1]))) == %(expected_filename)r,\
'Error, cannot launch extension module %(module_name)s, original package is in the way.'"""
    else:
        python_command_template = """\
import os, importlib.util;\
assert os.path.normcase(os.path.abspath(os.path.normpath(\
importlib.util.find_spec('%(module_name)s').origin))) == %(expected_filename)r,\
'Error, cannot launch extension module %(module_name)s, original package is in the way.'"""

    output_dir = os.path.normpath(getOutputDir())
    if output_dir != ".":
        python_command_template = """\
import sys; sys.path.insert(0, %(output_dir)r)
""" + python_command_template

    python_command_template += ";__import__('%(module_name)s')"

    python_command = python_command_template % {
        "module_name": tree.getName(),
        "expected_filename": os.path.normcase(
            os.path.abspath(
                os.path.normpath(
                    OutputDirectories.getResultFullpath(onefile=False, real=True)
                )
            )
        ),
        "output_dir": output_dir,
    }

    if shallRunInDebugger():
        args = wrapCommandForDebuggerForExec(
            command=(sys.executable, "-c", python_command),
            debugger=getDebuggerName(),
        )
    else:
        args = (sys.executable, "python", "-c", python_command)

    callExecPython(args=args, add_path=True, uac=False)


def compileTree():
    source_dir = OutputDirectories.getSourceDirectoryPath(onefile=False, create=False)

    general.info("Completed Python level compilation and optimization.")

    if not shallOnlyExecCCompilerCall():
        general.info("Generating source code for C backend compiler.")

        reportMemoryUsage(
            "before_c_code_generation",
            (
                "Total memory usage before generating C code"
                if isShowProgress() or isShowMemory()
                else None
            ),
        )

        # Now build the target language code for the whole tree.
        with withProfiling(
            name="code-generation",
            logger=code_generation_logger,
            enabled=isCompileTimeProfile(),
        ):
            module_filenames = makeSourceDirectory()

        bytecode_accessor = ConstantAccessor(
            data_filename="__bytecode.const", top_level_name="bytecode_data"
        )

        # This should take all bytecode values, even ones needed for frozen or
        # not produce anything.
        loader_code = LoaderCodes.getMetaPathLoaderBodyCode(bytecode_accessor)

        writeSourceCode(
            filename=getNormalizedPathJoin(source_dir, "__loader.c"),
            source_code=loader_code,
            logger=code_generation_logger,
            assume_yes_for_downloads=assumeYesForDownloads(),
        )

    else:
        # Too hard to come by this information, and it's only for optional parts
        # of the reporting at this time.
        module_filenames = {}

        source_dir = OutputDirectories.getSourceDirectoryPath(
            onefile=False, create=False
        )

        if not os.path.isfile(os.path.join(source_dir, "__helpers.h")):
            return general.sysexit("Error, no previous build directory exists.")

    reportMemoryUsage(
        "before_running_scons",
        (
            "Total memory usage before running scons"
            if isShowProgress() or isShowMemory()
            else None
        ),
    )

    if isShowMemory():
        printInstanceCounterStats()

    Reports.doMissingOptimizationReport()

    if shallNotDoExecCCompilerCall():
        return True, {}

    general.info("Running data composer tool for optimal constant value handling.")

    runDataComposer(source_dir)  # TODO: This should be a hook too

    writeExtraCodeFiles(onefile=False)

    provideStaticSourceFilesBackend(source_dir=source_dir)

    onGeneratedSourceCode(source_dir, onefile=False)

    general.info("Running C compilation via Scons.")

    # Run the Scons to build things.
    result, scons_options = runSconsBackend()

    if result:
        readSconsObjectSizes(
            source_dir=source_dir,
            module_filenames=module_filenames,
            module_mode=shallMakeModule(),
        )

    return result, scons_options


def handleSyntaxError(e):
    # Syntax or indentation errors, output them to the user and abort. If
    # we are not in full compat, and user has not specified the Python
    # versions he wants, tell him about the potential version problem.
    error_message = SyntaxErrors.formatOutput(e)

    if not states.is_full_compat:
        suggested_python_version_str = getSupportedPythonVersions()[-1]

        error_message += """

Nuitka is very syntax compatible with standard Python. It is currently running
with Python version '%s', you might want to specify more clearly with the use
of the precise Python interpreter binary and '-m nuitka', e.g. use this
'python%s -m nuitka' option, if that's not the one the program expects.
""" % (
            python_version_str,
            suggested_python_version_str,
        )

    # Important to have the same error
    sys.exit(error_message)


def _considerPgoInput():
    pgo_filename = getPythonPgoInput()
    if pgo_filename is not None:
        readPGOInputFile(pgo_filename)


def _detectEarlyDLLs():
    assert isStandaloneMode()
    modules = ModuleRegistry.getDoneModules()

    for module in modules:
        if module.isPythonExtensionModule():
            addExtensionModuleEntryPoint(module)

        addIncludedEntryPoints(considerExtraDlls(module))

    detectUsedDLLs(
        standalone_entry_points=getStandaloneEntryPoints(),
        source_dir=OutputDirectories.getSourceDirectoryPath(
            onefile=False, create=False
        ),
    )


def _detectLateDLLs(scons_options):
    assert isStandaloneMode()
    binary_filename = scons_options["result_exe"]

    detectUsedDLLs(
        standalone_entry_points=(addMainEntryPoint(binary_filename),),
        source_dir=OutputDirectories.getSourceDirectoryPath(
            onefile=False, create=False
        ),
    )


def _main():
    """Main program flow of Nuitka

    At this point, options will be parsed already, Nuitka will be executing
    in the desired version of Python with desired flags, and we just get
    to execute the task assigned.

    We might be asked to only re-compile generated C, dump only an XML
    representation of the internal node tree after optimization, etc.
    """

    # Main has to fulfill many options, leading to many branches and statements
    # to deal with them.  pylint: disable=too-many-branches,too-many-statements

    general.info(
        leader="Starting Python compilation with:",
        message="%s %s %s."
        % doNotBreakSpaces(
            "Version '%s'" % getNuitkaVersion(),
            "on Python %s (flavor '%s'%s)"
            % (
                python_version_str,
                getPythonFlavorName(),
                ("" if isPythonWithGil() else " no GIL"),
            ),
            "commercial grade '%s'" % (getCommercialVersion() or "not installed"),
        ),
    )

    # In case we are in a PGO run, we read its information first, so it becomes
    # available for later parts.
    _considerPgoInput()

    reportMemoryUsage(
        "after_launch",
        (
            "Total memory usage before processing"
            if isShowProgress() or isShowMemory()
            else None
        ),
    )

    # Initialize the importing layer from options, main filenames, debugging
    # options, etc.
    setupImportingFromOptions()

    # Initialize value tracing if requested.
    setupValueTraceFromOptions()

    # Let the plugins know we are starting compilation and they should make their checks.
    onCompilationStartChecks()

    addIncludedDataFilesFromFlavor()
    addIncludedDataFilesFromFileOptions()
    addIncludedDataFilesFromPackageOptions()

    # Turn that source code into a node tree structure.
    try:
        main_module = _createMainModule()
    except (SyntaxError, IndentationError) as e:
        return handleSyntaxError(e)

    addIncludedDataFilesFromMacOSAppOptions()
    addIncludedDataFilesFromPlugins()

    # Check if the --project expectations are met for data files.
    checkProjectExpectedDataFiles()

    dumpTreeXML()

    # Detect DLLs used so far.
    if isStandaloneMode():
        _detectEarlyDLLs()

    # Make the actual C compilation.
    result, scons_options = compileTree()

    # Exit if compilation failed.
    if not result:
        general.sysexit(
            message="Failed unexpectedly in Scons C backend compilation.",
            mnemonic="scons-backend-failure",
            reporting=True,
        )

    # Relaunch in case of Python PGO input to be produced.
    if shallCreatePythonPgoInput():
        # Will not return.
        pgo_filename = OutputDirectories.getPgoRunInputFilename()
        general.info(
            "Restarting compilation using collected information from '%s'."
            % pgo_filename
        )
        reExecuteNuitka(pgo_filename=pgo_filename)

    if shallNotDoExecCCompilerCall():
        if isShowMemory():
            showMemoryTrace()

        printPluginUsageStats()

        sys.exit(0)

    executePostProcessing(scons_options["result_exe"])

    if isStandaloneMode():
        _detectLateDLLs(scons_options)

        dist_dir = OutputDirectories.getStandaloneDirectoryPath(bundle=True, real=False)

        if not shallOnlyExecCCompilerCall():
            main_standalone_entry_point, copy_standalone_entry_points = copyDllsUsed(
                dist_dir=dist_dir,
                standalone_entry_points=getStandaloneEntryPoints(),
            )

            data_file_paths = copyDataFiles(
                standalone_entry_points=getStandaloneEntryPoints()
            )

            if isMacOS():
                signDistributionMacOS(
                    dist_dir=dist_dir,
                    data_file_paths=data_file_paths,
                    main_standalone_entry_point=main_standalone_entry_point,
                    copy_standalone_entry_points=copy_standalone_entry_points,
                )

            dist_dir = OutputDirectories.renameStandaloneDirectory(dist_dir)

        onStandaloneDistributionFinished(
            dist_dir=dist_dir,
            standalone_binary=OutputDirectories.getResultFullpath(
                onefile=False, real=True
            ),
        )

        if isOnefileMode():
            packDistFolderToOnefile(dist_dir)

            if isRemoveBuildDir():
                general.info("Removing dist folder '%s'." % dist_dir)

                removeDirectory(
                    path=dist_dir,
                    logger=general,
                    ignore_errors=False,
                    extra_recommendation=None,
                )
            else:
                general.info(
                    "Keeping dist folder '%s' for inspection, no need to use it."
                    % dist_dir
                )

    # Remove the source directory (now build directory too) if asked to.
    source_dir = OutputDirectories.getSourceDirectoryPath(onefile=False, create=False)

    if isRemoveBuildDir():
        general.info("Removing build directory '%s'." % source_dir)

        # Make sure the scons report is cached before deleting it.
        readSconsReport(source_dir)
        readSconsErrorReport(source_dir)
        readSconsResourceUsageReports(source_dir)

        removeDirectory(
            path=source_dir,
            logger=general,
            ignore_errors=False,
            extra_recommendation=None,
        )
        assert not os.path.exists(source_dir)
    else:
        general.info("Keeping build directory '%s'." % source_dir)

    final_filename = OutputDirectories.getResultFullpath(
        onefile=isOnefileMode(), real=True
    )

    if isStandaloneMode() and isMacOS():
        general.info(
            "Created binary that runs on macOS %s (%s) or higher."
            % (scons_options["macos_min_version"], scons_options["macos_target_arch"])
        )

        if scons_options["macos_target_arch"] != getArchitecture():
            general.warning(
                "It will only work as well as 'arch -%s %s %s' does."
                % (
                    scons_options["macos_target_arch"],
                    sys.executable,
                    getMainEntryPointFilenames()[0],
                ),
                mnemonic="macos-cross-compile",
            )

    onFinalResult(final_filename)

    if shallMakeModule():
        base_path = OutputDirectories.getResultBasePath(onefile=False)

        if os.path.isdir(base_path) and getPackageDirFilename(base_path):
            general.warning(
                """\
The compilation result is hidden by package directory '%s'. Importing will \
not use compiled code while it exists because it has precedence while both \
exist, out e.g. '--output-dir=output' to sure is importable.""" % base_path,
                mnemonic="compiled-package-hidden-by-package",
            )

    general.info("Successfully created '%s'." % getReportPath(final_filename))

    # Archive creations, installer creations go here.
    if shallCreateDmgFile():
        createDmgFile(general)

    writeCompilationReports(aborted=False)
    printPluginUsageStats()

    run_filename = OutputDirectories.getResultRunFilename(onefile=isOnefileMode())

    # Execute the module immediately if option was given.
    if shallExecuteImmediately():
        general.info("Launching '%s'." % run_filename)

        if shallMakeModule():
            _executeModule(tree=main_module)
        else:
            _executeMain(run_filename)
    else:
        if run_filename != final_filename:
            general.info(
                "Execute it by launching '%s', the batch file needs to set environment."
                % run_filename
            )


def main():
    try:
        _main()
    except BaseException:
        try:
            writeCompilationReports(aborted=True)
        except KeyboardInterrupt:
            general.warning("""Report writing was prevented by user interrupt.""")
        # Catch all the things, pylint: disable-next=broad-exception-caught
        except BaseException as e:
            general.warning("""\
Report writing was prevented by exception %r, use option \
'--experimental=debug-report-traceback' for full traceback.""" % e)

            if isExperimental("debug-report-traceback"):
                raise

        raise


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#!/usr/bin/env python
#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Main program for yamllint checker tool."""

import os
import sys

# Unchanged, running from checkout, use the parent directory, the nuitka
# package ought be there.
sys.path.insert(
    0, os.path.normpath(os.path.join(os.path.dirname(__file__), "..", "..", ".."))
)

# isort:start

from nuitka.format.YamlChecker import checkYamlSchema
from nuitka.options.CommandLineOptionsTools import makeOptionsParser
from nuitka.tools.Basics import goHome
from nuitka.tools.quality.ScanSources import scanTargets
from nuitka.Tracing import my_print, tools_logger
from nuitka.utils.FileOperations import resolveShellPatternToFilenames


def main():
    parser = makeOptionsParser(usage="%prog [options]", epilog=None)

    parser.add_option(
        "--update-checksum",
        dest="update_checksum",
        action="store_true",
        default=False,
        help="""\
Update the version checksum after checking, so Nuitka knowns it can be trusted.""",
    )

    options, positional_args = parser.parse_args()

    if not positional_args:
        positional_args = [
            "nuitka/plugins/standard/*.yml",
            "nuitka/plugins/commercial/*.yml",
        ]

    my_print("Working on:", positional_args)

    positional_args = sum(
        (
            resolveShellPatternToFilenames(positional_arg)
            for positional_arg in positional_args
        ),
        [],
    )

    goHome()

    filenames = list(
        scanTargets(
            positional_args,
            suffixes=(".yaml",),
        )
    )
    if not filenames:
        tools_logger.sysexit("No files found.")

    for filename in filenames:
        checkYamlSchema(
            filename=filename,
            effective_filename=filename,
            update=options.update_checksum,
            logger=tools_logger,
        )


if __name__ == "__main__":
    main()

#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Tests for coroutine throw() and cancel/await bridge handling."""

import asyncio
import sys


class AwaitException(Exception):
    pass


throw_namespace = {"AwaitException": AwaitException}

exec(
    """\
import types

@types.coroutine
def awaitable():
    yield ("throw",)

async def leaf():
    try:
        await awaitable()
    except AwaitException:
        return b"payload"

async def mid():
    body = await leaf()
    return {"body": body}

async def leaf_empty():
    try:
        await awaitable()
    except AwaitException:
        return

async def mid_empty():
    body = await leaf_empty()
    return {"body": body}
""",
    throw_namespace,
)

mid = throw_namespace["mid"]
mid_empty = throw_namespace["mid_empty"]


def run_until_complete(coro):
    exc = False

    while True:
        try:
            if exc:
                exc = False
                fut = coro.throw(AwaitException)
            else:
                fut = coro.send(None)
        except StopIteration as ex:
            return ex.args[0] if ex.args else None

        if fut == ("throw",):
            exc = True


async def top():
    return await mid()


async def top_empty():
    return await mid_empty()


print(run_until_complete(top()))
print(run_until_complete(top_empty()))

send_namespace = {}

exec(
    """\
import types

@types.coroutine
def resume_with_value():
    result = yield ("send",)
    return result
""",
    send_namespace,
)


def run_until_complete_with_send(coro, value):
    first = coro.send(None)
    assert first == ("send",), first

    try:
        coro.send(value)
    except StopIteration as ex:
        return ex.args[0] if ex.args else None

    assert False, "coroutine should have completed"


async def top_send_value_bridge():
    result = await send_namespace["resume_with_value"]()
    print(type(result).__module__, type(result).__name__, repr(result))


run_until_complete_with_send(top_send_value_bridge(), {"sent": 17})

# Keep the inner coroutine uncompiled so the await path crosses the
# compiled/uncompiled coroutine boundary during cancellation handling.
bridge_namespace = {}

exec(
    """\
import asyncio

def get_current_task(loop):
    if hasattr(asyncio, "current_task"):
        return asyncio.current_task()

    return asyncio.Task.current_task(loop=loop)

def create_future(loop):
    if hasattr(loop, "create_future"):
        return loop.create_future()

    return asyncio.Future(loop=loop)

def create_task(loop, coro):
    if hasattr(asyncio, "create_task"):
        return asyncio.create_task(coro)

    return asyncio.ensure_future(coro, loop=loop)

async def run_inner():
    loop = asyncio.get_event_loop()
    current_task = get_current_task(loop)
    result_future = create_future(loop)

    async def request_cancel():
        current_task.cancel()

    cancel_task = create_task(loop, request_cancel())

    def deliver_result(_task):
        if not result_future.done():
            result_future.set_result(42)

    cancel_task.add_done_callback(deliver_result)

    try:
        await asyncio.sleep(3600)
    except asyncio.CancelledError:
        return await result_future
""",
    bridge_namespace,
)


async def top_cancel_await_bridge():
    result = await bridge_namespace["run_inner"]()
    print(type(result).__module__, type(result).__name__, repr(result))


def run_async(coro):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    try:
        return loop.run_until_complete(coro)
    finally:
        asyncio.set_event_loop(None)
        loop.close()


run_async(top_cancel_await_bridge())

#     Python tests originally created or extracted from other peoples work. The
#     parts were too small to be protected.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""PyLint handling for Nuitka.

Our usage of PyLint also works around a few issues that PyLint
has.

"""

import os
import sys

from nuitka.tools.testing.Common import hasModule, my_print
from nuitka.utils.Execution import check_output, executeProcess, getNullOutput
from nuitka.utils.FileOperations import getNormalizedPath

_pylint_version = None


def checkVersion():
    # pylint: disable=global-statement
    global _pylint_version

    if not hasModule("pylint"):
        sys.exit(
            "Error, pylint is not installed for this interpreter '%s' version."
            % os.environ["PYTHON"]
        )

    if _pylint_version is None:
        with getNullOutput() as null_output:
            _pylint_version = check_output(
                [os.environ["PYTHON"], "-m", "pylint", "--version"], stderr=null_output
            )

        if str is not bytes:
            _pylint_version = _pylint_version.decode("utf8")

        _pylint_version = _pylint_version.split("\n")[0].split()[-1].strip(",")

    my_print("Using PyLint version:", _pylint_version)

    return tuple(int(d) for d in _pylint_version.split("."))


# Disabled globally:
#
# no-init: Class has no __init__ method
# Who cares, we are using overrides that don't need to change object init a lot
# and rarely ever made a mistake with forgetting to call used __init__ of the
# parent.
#
# I0011: Locally disabling W....
# Strange one anyway, we want to locally disable stuff. And that just makes it
# a different warning. Amazing. Luckily we can decide to ignore that globally
# then.
#
# I0012: Locally enabling W....
# Sure, we disabled it for a block, and re-enabled it then.
#
# bad-whitespace: No space allowed...
# Our spaces before keyword argument calls are not allowed, and this is
# not possible to distinguish.
#
# bad-continuation: Wrong hanging line indentation
# No it's not wrong.
#
# E1120 / E1123: ....
# Constructor call checks frequently fail miserably, so this is full of
# mysterious false alarms, while it's unlikely to help much.
#
# E1103: Instance of 'x' has no 'y' member but some types could not be inferred
# Rarely is this any help, but it's full of false alarms.
#
# W0632: Possible unbalanced tuple unpacking with sequence defined at ...
# It's not really good at guessing these things.
#
# W1504: Using type() instead of isinstance() for typechecking.
# Nuitka is all about exact type checks, so this doesn't apply
#
# C0123: Using type() instead of isinstance() for typechecking.
# Nuitka is all about exact type checks, so this doesn't apply
#
# C0413: Import "..." should be placed at the top of the module
# There is no harm to this and imports are deal with by isort binary.
#
# C0411: external import "external" comes before "local"
# There is no harm to this and imports are deal with by "isort" binary.
#
# R0204: Redefinition of var type from x to y
# We do this all the time, e.g. to convert "str" to "unicode", or "list" to "str".
#
# R1705: Unnecessary "else" after "return"
# Frequently we use multiple branches where each returns.
#
# inconsistent-return-statements
# This makes no sense, having to have a variable for return is bad in my mind.
#
# c-extension-no-member
# Not too useful for us.
#
# useless-object-inheritance
# The code is for Python2 still, where it makes a difference, if you do
# not specify a base class, object is not the default there, but old style
# classes are, which perform different/worse.
#
# useless-return
# We like explicit None returns where the return value can be overloaded
# to something else, or the function is used along others that do return
# other things.
#
# ungrouped-imports
# We let isort do its thing most of the time, and where not, it's good
# enough for us to handle it manually.
#
# assignment-from-no-return
# assignment-from-none
# Overloaded functions are not detected, default value returns are all
# warned about, not worth it.
#
# raise-missing-from
# cannot do that, as long as we are backwards compatible

# import-outside-toplevel
# We do this deliberately, to avoid importing modules we do not use in
# all cases, e.g. Windows/macOS specific stuff.

# consider-using-f-string
# We need to be backward compatible for Python versions that do not have
# it.

# super-with-arguments
# Keeping code portable to Python2 is still good.

# consider-using-dict-comprehension
# Keeping code portable to Python2 is still good.

# unnecessary-lambda-assignment
# For deciders, we do this and like it.

# unnecessary-dunder-call
# We do make those intentionally only.

# arguments-differ
# We override static methods with non-static all the time.

# use-yield-from
# Keeping code portable to Python2 is still good.

# unnecessary-lambda-assignment
# For deciders, we do this and like it.

# unnecessary-dunder-call
# We do make those intentionally only.

# arguments-differ
# Not sure

# redefined-slots-in-subclass
# We do this to allow for bases classes and mixins and different slot loadout.


def getOptions():
    pylint_version = checkVersion()

    # spell-checker: ignore setrecursionlimit,rcfile

    default_pylint_options = """\
--init-hook=import sys;sys.setrecursionlimit(1024*sys.getrecursionlimit())
--disable=I0011,E1103,W0632,\
C0123,C0411,C0413,cyclic-import,duplicate-code,\
deprecated-module,deprecated-method,deprecated-argument,assignment-from-none,\
ungrouped-imports,no-else-return,c-extension-no-member,\
inconsistent-return-statements,raise-missing-from,import-outside-toplevel,\
useless-object-inheritance,useless-return,assignment-from-no-return,\
redundant-u-string-prefix,consider-using-f-string,consider-using-dict-comprehension,\
unnecessary-lambda-assignment,unnecessary-dunder-call,arguments-differ,\
redefined-slots-in-subclass
--enable=useless-suppression
--msg-template="{path}:{line} {msg_id} {symbol} {obj} {msg}"
--reports=no
--persistent=no
--method-rgx=[a-z_][a-zA-Z0-9_]{2,55}$
--module-rgx=.*
--function-rgx=.*
--variable-rgx=.*
--argument-rgx=.*
--dummy-variables-rgx=_.*|trace_collection
--ignored-argument-names=_.*|trace_collection
--const-rgx=.*
--max-line-length=125
--no-docstring-rgx=.*
--max-module-lines=6000
--min-public-methods=0
--max-public-methods=200
--max-args=11
--max-parents=17
--max-statements=50
--max-nested-blocks=10
--max-bool-expr=10
--load-plugins=pylint.extensions.no_self_use
--score=no\
""".split("\n")

    if os.name != "nt":
        default_pylint_options.append("--rcfile=%s" % os.devnull)

    if pylint_version >= (4, 0):
        default_pylint_options.append("--disable=use-yield-from")
        default_pylint_options.append("--max-positional-arguments=20")
        default_pylint_options.append("--disable=deprecated-class")
    return default_pylint_options


our_exit_code = 0


def _cleanupPylintOutput(output):
    if str is not bytes:
        output = output.decode("utf8")

    # Normalize from Windows newlines potentially
    output = output.replace("\r\n", "\n")

    lines = [
        line
        for line in output.split("\n")
        if line
        if "Using config file" not in line
        if "Unable to import 'resource'" not in line
        if "Bad option value 'self-assigning-variable'" not in line
    ]

    try:
        error_line = lines.index("No config file found, using default configuration")
        del lines[error_line]

        if error_line < len(lines):
            del lines[error_line]
    except ValueError:
        pass

    return lines


def _executePylint(filenames, pylint_options, extra_options):
    # This is kind of a singleton module, pylint: disable=global-statement
    global our_exit_code

    command = (
        [os.environ["PYTHON"], "-m", "pylint"]
        + pylint_options
        + extra_options
        + filenames
    )

    env = dict(os.environ)
    env["PYTHONWARNINGS"] = "ignore"

    process_result = executeProcess(
        command,
        env=env,
    )

    if process_result.exit_code == -11:
        sys.exit("Error, segfault from pylint.")

    # Catch random crashes with non standard exit code.
    if process_result.exit_code < 0 or process_result.exit_code >= 64:
        sys.exit(
            "Error, strange crash with exit_code %d from pylint."
            % process_result.exit_code
        )

    stdout = _cleanupPylintOutput(process_result.stdout)
    stderr = _cleanupPylintOutput(process_result.stderr)

    if stderr:
        our_exit_code = 1

        for line in stderr:
            my_print(line)

    if stdout:
        # If we filtered everything away, remove the leading file name reports.
        while stdout and stdout[-1].startswith("******"):
            del stdout[-1]

        for line in stdout:
            my_print(line)

        if stdout:
            our_exit_code = 1

    sys.stdout.flush()


def hasPyLintBugTrigger(filename):
    """Decide if a filename should be skipped."""
    # Currently everything is good, but it's a useful hook, pylint_: disable=unused-argument
    if getNormalizedPath(filename) == getNormalizedPath("nuitka/distutils/Build.py"):
        return True

    return False


def isSpecificPythonOnly(filename):
    """Decide if something is not used for this specific Python."""

    # Currently everything is portable, but it's a useful hook, pylint: disable=unused-argument
    return False


def executePyLint(filenames, show_todo, verbose, one_by_one):
    filenames = list(filenames)

    if verbose:
        my_print("Checking", filenames, "...")

    pylint_options = getOptions()
    if not show_todo:
        pylint_options.append("--notes=")

    filenames = [
        filename
        for filename in filenames
        if not hasPyLintBugTrigger(filename)
        if not isSpecificPythonOnly(filename)
    ]

    extra_options = os.getenv("PYLINT_EXTRA_OPTIONS", "").split()
    if "" in extra_options:
        extra_options.remove("")

    if one_by_one:
        for filename in filenames:
            my_print("Checking", filename, ":")
            _executePylint([filename], pylint_options, extra_options)
    else:
        _executePylint(filenames, pylint_options, extra_options)

    return our_exit_code


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Plugins usage statistics"""

from nuitka.States import states
from nuitka.Tracing import printIndented, printLine
from nuitka.utils.Timing import TimerReport

counted_plugin_methods = {}


def counted_plugin_method(plugin_method):
    name = "Plugins." + plugin_method.__name__

    def wrapped_plugin_method(*args, **kw):
        if states.show_plugin_usage:
            if name not in counted_plugin_methods:
                counted_plugin_methods[name] = [0, 0.0]

            counted_plugin_methods[name][0] += 1

            timer_report = TimerReport(
                message="", decider=False, include_sleep_time=False
            )
            with timer_report:
                result = plugin_method(*args, **kw)

            counted_plugin_methods[name][1] += timer_report.getTimer().getDelta()

            return result

        return plugin_method(*args, **kw)

    return wrapped_plugin_method


def printPluginUsageStats():
    if not states.show_plugin_usage:
        return

    if counted_plugin_methods:
        printLine("Plugin method calls:")

        for name, (count, total_time) in sorted(
            counted_plugin_methods.items(), key=lambda x: x[1][1], reverse=True
        ):
            average_time = total_time / count if count > 0 else 0.0
            printIndented(
                1,
                "%s calls: %d, total time: %.3fs, avg time: %.5fs"
                % (name, count, total_time, average_time),
            )


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Plugin for including selected source code recorded from probe execution."""

import json
import os
import shlex
import sys

from nuitka.containers.OrderedDicts import OrderedDict
from nuitka.importing.FakeModules import makeFakeModuleDescription
from nuitka.options.Options import (
    getMainEntryPointFilenames,
    shallMakeDll,
    shallMakeModule,
)
from nuitka.plugins.PluginBase import NuitkaPluginBase
from nuitka.utils.Execution import executeProcess
from nuitka.utils.ModuleNames import ModuleName


def _decodeCommandOutput(value):
    if str is not bytes:
        return value.decode("utf8", "replace")
    else:
        return value


def _joinSortedKeys(value):
    return ",".join(sorted(value))


def _formatDictLiteralCode(value):
    if not value:
        return "{}"

    result = ["{"]

    for key in value:
        result.append("    %r: %r," % (key, value[key]))

    result.append("}")

    return "\n".join(result)


_probe_result_prefix = "NUITKA_SOURCE_INCLUSION_JSON:"


class NuitkaPluginSourceInclusion(NuitkaPluginBase):
    plugin_name = "source-inclusion"
    plugin_desc = "Record and replay selected 'inspect.getsource()' lookups."
    plugin_category = "feature,package-support"

    @classmethod
    def addPluginCommandLineOptions(cls, group):
        group.add_option(
            "--source-inclusion-args",
            action="append",
            dest="source_inclusion_arg_sets",
            default=[],
            help="""\
Run the program being compiled during compilation with this shell-style
argument list while recording successful 'inspect.getsource()' lookups.
Repeat for every additional probe run, e.g.
'--source-inclusion-args="subcommand --help"'.""",
        )
        group.add_option(
            "--source-inclusion-timeout",
            action="store",
            dest="source_inclusion_timeout",
            type="int",
            default=30,
            help="""\
Timeout in seconds for each probe run. Defaults to 30.""",
        )
        group.add_option(
            "--show-source-inclusion-trace",
            action="store_true",
            dest="show_source_inclusion_trace",
            default=False,
            help="""\
Display the recorded 'inspect.getsource()' keys from the probe run.""",
        )

    @classmethod
    def isRelevant(cls):
        return not (shallMakeModule() or shallMakeDll())

    def _normalizeSourceInclusionArgSets(self, source_inclusion_arg_sets):
        result = []

        for source_inclusion_arg_set in source_inclusion_arg_sets:
            try:
                result.append(tuple(shlex.split(source_inclusion_arg_set)))
            except ValueError as e:
                return self.sysexit(
                    "Error, malformed value %r for '--source-inclusion-args': %s"
                    % (source_inclusion_arg_set, e)
                )

        if not result:
            result.append(())

        return tuple(result)

    def __init__(
        self,
        source_inclusion_arg_sets,
        source_inclusion_timeout,
        show_source_inclusion_trace,
    ):
        if source_inclusion_timeout <= 0:
            self.sysexit(
                "Error, '--source-inclusion-timeout' must be greater than zero."
            )

        self.source_inclusion_arg_sets = self._normalizeSourceInclusionArgSets(
            source_inclusion_arg_sets=source_inclusion_arg_sets
        )

        self.source_inclusion_timeout = source_inclusion_timeout
        self.show_source_inclusion_trace = show_source_inclusion_trace

        self.main_filename = None
        self.source_data = OrderedDict()

    def onCompilationStartChecks(self):
        main_filenames = getMainEntryPointFilenames()

        if len(main_filenames) != 1:
            return self.sysexit(
                "Error, '--enable-plugin=source-inclusion' does not support multiple main programs yet."
            )

        self.main_filename = main_filenames[0]

        source_data = self._runProbeExecutions(main_filename=self.main_filename)

        self.source_data = source_data

        self.info(
            "Recorded %d successful 'inspect.getsource()' lookups from %d probe execution(s)."
            % (len(self.source_data), len(self.source_inclusion_arg_sets))
        )

        if self.show_source_inclusion_trace:
            if self.source_data:
                self.info(
                    "Recorded source inclusion keys: %s"
                    % _joinSortedKeys(self.source_data)
                )
            else:
                self.info("Recorded source inclusion keys: none")

        if not self.source_data:
            self.warning("""\
No successful 'inspect.getsource()' lookups were recorded. Adjust \
'--source-inclusion-args' values if source inspection still fails at run \
time.""")

    def _runProbeExecutions(self, main_filename):
        merged_source_data = {}

        for probe_args in self.source_inclusion_arg_sets:
            source_data = self._runProbeExecution(
                main_filename=main_filename, probe_args=probe_args
            )

            for key, source_code in source_data.items():
                if key in merged_source_data:
                    if merged_source_data[key] != source_code:
                        return self.sysexit(
                            "Error, probe execution for '%s' with arguments %r recorded conflicting source code for %r."
                            % (main_filename, probe_args, key)
                        )
                else:
                    merged_source_data[key] = source_code

        ordered_source_data = OrderedDict()

        for key in sorted(merged_source_data):
            ordered_source_data[key] = merged_source_data[key]

        return ordered_source_data

    def _runProbeExecution(self, main_filename, probe_args):
        probe_code = self._getProbeCode()

        command = [sys.executable, "-c", probe_code, main_filename]
        command += list(probe_args)

        env = dict(os.environ)
        env["PYTHONIOENCODING"] = "utf8"

        process_result = executeProcess(
            command=command,
            env=env,
            stdin=False,
            timeout=self.source_inclusion_timeout,
        )

        stdout = _decodeCommandOutput(process_result.stdout).strip()
        stderr = _decodeCommandOutput(process_result.stderr).strip()

        if process_result.exit_code != 0:
            return self.sysexit(
                "Error, probe execution for '%s' with arguments %r failed with exit code %d.\nStdout: %s\nStderr: %s"
                % (
                    main_filename,
                    probe_args,
                    process_result.exit_code,
                    stdout or "<empty>",
                    stderr or "<empty>",
                )
            )

        return self._decodeProbeResult(main_filename=main_filename, stdout=stdout)

    def _decodeProbeResult(self, main_filename, stdout):
        stdout_lines = [line for line in stdout.splitlines() if line.strip()]

        if not stdout_lines:
            return self.sysexit(
                "Error, probe execution for '%s' produced no JSON result."
                % main_filename
            )

        for stdout_line in reversed(stdout_lines):
            if stdout_line.startswith(_probe_result_prefix):
                probe_result_text = stdout_line[len(_probe_result_prefix) :]
                break
        else:
            return self.sysexit(
                "Error, probe execution for '%s' did not produce a tagged JSON result."
                % main_filename
            )

        try:
            probe_result = json.loads(probe_result_text)
        except ValueError:
            return self.sysexit(
                "Error, probe execution for '%s' did not produce valid JSON: %s"
                % (main_filename, probe_result_text)
            )

        source_data = probe_result.get("sources", {})

        if type(source_data) is not dict:
            return self.sysexit(
                "Error, probe execution for '%s' returned invalid source data."
                % main_filename
            )

        ordered_source_data = OrderedDict()

        for key in sorted(source_data):
            ordered_source_data[key] = source_data[key]

        return ordered_source_data

    @staticmethod
    def _getProbeCode():
        return r"""
from __future__ import absolute_import
from __future__ import print_function

import inspect
import io
import json
import os
import runpy
import sys


def _getSourceInclusionKey(obj):
    func = getattr(obj, "__func__", None)

    if func is not None and not inspect.isclass(obj):
        obj = func

    if inspect.ismodule(obj):
        module_name = getattr(obj, "__name__", None)

        if module_name is not None:
            return "module:" + module_name

    module_name = getattr(obj, "__module__", None)
    qualname = getattr(obj, "__qualname__", None)

    if qualname is None:
        qualname = getattr(obj, "__name__", None)

    if module_name is not None and qualname is not None:
        return "object:" + module_name + ":" + qualname

    return None


def _main():
    main_filename = os.path.abspath(sys.argv[1])
    probe_args = sys.argv[2:]

    source_data = {}

    original_getsource = inspect.getsource

    def _recording_getsource(obj):
        key = _getSourceInclusionKey(obj)

        try:
            result = original_getsource(obj)
        except Exception:
            raise
        else:
            if key is not None and key not in source_data:
                source_data[key] = result

            return result

    inspect.getsource = _recording_getsource

    null_output = io.open(os.devnull, "w", encoding="utf8", errors="ignore")
    old_stdout = sys.stdout
    old_stderr = sys.stderr

    try:
        sys.stdout = null_output
        sys.stderr = null_output
        sys.argv = [main_filename] + probe_args

        main_directory = os.path.dirname(main_filename)
        if main_directory:
            sys.path[0] = main_directory

        try:
            runpy.run_path(main_filename, run_name="__main__")
        except SystemExit as e:
            if e.code not in (None, 0):
                raise
    finally:
        sys.stdout = old_stdout
        sys.stderr = old_stderr
        null_output.close()

    print(
        "NUITKA_SOURCE_INCLUSION_JSON:%s"
        % json.dumps({"sources": source_data}, sort_keys=True)
    )


_main()
""".strip()

    def createFakeModuleDependency(self, module):
        if module.getFullName() == "inspect" and self.source_data:
            return (
                makeFakeModuleDescription(
                    module_name=ModuleName("_nuitka_source_inclusion"),
                    source_code=self._getSupportModuleCode(),
                    source_filename=self.main_filename,
                    reason="recorded source inclusion support",
                ),
            )

    def createPostModuleLoadCode(self, module):
        if module.getFullName() == "inspect" and self.source_data:
            return (
                """\
import inspect
import _nuitka_source_inclusion
_nuitka_source_inclusion.install(inspect)
""",
                "Install recorded source inclusion for 'inspect.getsource()'.",
            )

    def _getSupportModuleCode(self):
        source_data_code = _formatDictLiteralCode(self.source_data)

        return """\
import inspect

_source_inclusion_data = %s


def _getSourceInclusionKey(obj):
    func = getattr(obj, "__func__", None)

    if func is not None and not inspect.isclass(obj):
        obj = func

    if inspect.ismodule(obj):
        module_name = getattr(obj, "__name__", None)

        if module_name is not None:
            return "module:" + module_name

    module_name = getattr(obj, "__module__", None)
    qualname = getattr(obj, "__qualname__", None)

    if qualname is None:
        qualname = getattr(obj, "__name__", None)

    if module_name is not None and qualname is not None:
        return "object:" + module_name + ":" + qualname

    return None


def install(inspect_module):
    if getattr(inspect_module, "_nuitka_source_inclusion_installed", False):
        return

    original_getsource = inspect_module.getsource

    def nuitka_getsource(obj):
        key = _getSourceInclusionKey(obj)

        if key is not None:
            source_code = _source_inclusion_data.get(key)

            if source_code is not None:
                return source_code

        return original_getsource(obj)

    inspect_module.getsource = nuitka_getsource
    inspect_module._nuitka_source_inclusion_installed = True
    inspect_module._nuitka_source_inclusion_sources = _source_inclusion_data
""" % source_data_code

    def getCacheContributionValues(self, module_name):
        del module_name

        yield self.plugin_name
        yield self.source_inclusion_arg_sets
        yield self.source_data

    def getReportData(self):
        return {
            "recorded_source_keys": tuple(self.source_data),
            "source_inclusion_arg_sets": self.source_inclusion_arg_sets,
            "source_inclusion_timeout": self.source_inclusion_timeout,
        }


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#!/usr/bin/env python
#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Tool to automatically format source code in Nuitka style."""

import contextlib
import os
import shutil

from nuitka.format.BiomeFormatter import formatJsonFile
from nuitka.format.FileFormatting import (
    cleanupTrailingWhitespace,
    cleanupWindowsNewlines,
    formatC,
)
from nuitka.format.PythonFormatting import formatPython
from nuitka.format.YamlChecker import checkYamlSchema
from nuitka.format.YamlFormatter import formatYaml
from nuitka.tools.quality.Git import (
    getFileHashContent,
    putFileHashContent,
    updateFileIndex,
    updateGitFile,
)
from nuitka.tools.quality.ScanSources import isPythonFile
from nuitka.tools.release.Documentation import extra_rst_keywords
from nuitka.Tracing import my_print, tools_logger
from nuitka.utils.Execution import check_call, getExecutablePath
from nuitka.utils.FileOperations import (
    addFileContentsBOM,
    deleteFile,
    getFileContentByLine,
    getFileContents,
    getFilenameExtension,
    openTextFile,
    stripFileContentsBOM,
    withPreserveFileMode,
    withTemporaryFile,
    withTemporaryFilename,
)
from nuitka.utils.PrivatePipSpace import (
    getMdformatBinaryPath,
    getRstfmtBinaryPath,
    withPrivatePipSitePackagesPathAdded,
)


def _shouldNotFormatCode(filename, effective_filename):
    """Check if a file should not be formatted."""
    # pylint: disable=too-many-return-statements

    parts = os.path.normpath(effective_filename).split(os.path.sep)

    if "inline_copy" in parts:
        if os.path.basename(effective_filename) == "scons.py":
            return False

        return True
    if "pybench" in parts:  # spell-checker: ignore pybench
        return True
    if "mercurial" in parts:
        return True
    if "tests" in parts and parts[parts.index("tests") + 1].startswith("CPython"):
        return True
    if "tests" in parts and "syntax" in parts:
        return True
    if "tests" in parts and "scratch" in parts:
        return True
    if ".dist/" in effective_filename:
        return True
    if os.path.basename(effective_filename) in ("incbin.h", "hedley.h"):
        return True

    if effective_filename.endswith(".py") and os.path.exists(filename):
        for line in getFileContentByLine(filename, encoding="utf8"):
            if "# encoding: nuitka-protection" in line:
                return True

            break

    return False


def _transferBOM(source_filename, target_filename):
    """Transfer Byte Order Mark (BOM) from source to target file."""
    with open(source_filename, "rb") as f:
        source_code = f.read()
        source_code, bom = stripFileContentsBOM(source_code)

    if bom:
        with open(target_filename, "rb") as f:
            source_code = f.read()

        updated_source_code = addFileContentsBOM(source_code)

        if updated_source_code != source_code:
            with open(target_filename, "wb") as f:
                f.write(updated_source_code)


def cleanupMarkdownFmt(logger, filename, assume_yes_for_downloads):
    """Cleanup markdown formatting using mdformat.

    Args:
        filename: path to the file
        logger: logger to use
        assume_yes_for_downloads: bool
    """

    mdformat_path = getMdformatBinaryPath(
        logger=logger, assume_yes_for_downloads=assume_yes_for_downloads
    )

    if mdformat_path is None:
        if logger is not None:
            logger.warning("Need to accept mdformat download to format markdown files.")
        return

    with withPrivatePipSitePackagesPathAdded(logger=logger):
        check_call([mdformat_path, "--number", "--wrap=100", filename])


def cleanupRstFmt(logger, filename, effective_filename, assume_yes_for_downloads):
    """Cleanup reStructuredText formatting using rstfmt.

    Args:
        filename: path to the file
        effective_filename: filename to use for errors
        logger: logger to use
        assume_yes_for_downloads: bool
    """
    updated_contents = contents = getFileContents(filename, mode="rb")

    for keyword in extra_rst_keywords:
        updated_contents = updated_contents.replace(
            b".. %s::" % keyword, b".. raw:: %s" % keyword
        )

    if updated_contents != contents:
        with open(filename, "wb") as out_file:
            out_file.write(updated_contents)

    rstfmt_path = getRstfmtBinaryPath(
        logger=logger, assume_yes_for_downloads=assume_yes_for_downloads
    )

    if rstfmt_path is None:
        if logger is not None:
            logger.warning("Need to accept rstfmt download to format RST files.")
        return

    with withPrivatePipSitePackagesPathAdded(logger=logger):
        check_call([rstfmt_path, filename])

    cleanupWindowsNewlines(filename, effective_filename)

    contents = getFileContents(filename, mode="rb")

    updated_contents = contents.replace(b".. code:: sh\n", b".. code:: bash\n")

    for keyword in extra_rst_keywords:
        updated_contents = updated_contents.replace(
            b".. raw:: %s" % keyword, b".. %s::" % keyword
        )

    lines = []
    inside = False
    needs_empty = False

    for line in updated_contents.splitlines():
        if line.startswith(b"-"):
            if inside and needs_empty:
                lines.append(b"")

            inside = True
            needs_empty = True
            lines.append(line)
        elif inside and line == b"":
            needs_empty = False
            lines.append(line)
        elif inside and line.startswith(b"  "):
            needs_empty = True
            lines.append(line)
        else:
            inside = False
            lines.append(line)

    updated_contents = b"\n".join(lines) + b"\n"

    if updated_contents != contents:
        with open(filename, "wb") as out_file:
            out_file.write(updated_contents)


def cleanupPngImage(filename, logger):
    """Cleanup PNG image using optipng.

    Args:
        filename: path to the file
        logger: logger to use
    """
    _optipng_path = getExecutablePath("optipng")

    if _optipng_path:
        check_call([_optipng_path, "-quiet", "-o7", "-zm1-9", filename])
    elif logger is not None:
        logger.warning("Cannot find 'optipng' binary to compress PNG image")


def cleanupJpegImage(filename, logger):
    """Cleanup JPEG image using jpegoptim.

    Args:
        filename: path to the file
        logger: logger to use
    """
    # spell-checker: ignore jpegoptim
    _jpegoptim_path = getExecutablePath("jpegoptim")

    if _jpegoptim_path:
        check_call([_jpegoptim_path, "-q", filename])
    elif logger is not None:
        logger.warning("Cannot find 'jpegoptim' binary to compress JPEG image")


def formatText(
    logger,
    filename,
    effective_filename,
    ignore_yaml_diff=True,
    assume_yes_for_downloads=False,
):
    """Format text-like files (RST, MD, YAML, etc.).

    Args:
        filename: path to the file
        effective_filename: filename to use for errors
        logger: logger to use
        ignore_yaml_diff: bool - if YAML diffs should be ignored
        assume_yes_for_downloads: bool - if tools should be downloaded automatically
    """
    is_rst = effective_filename.endswith((".rst", ".inc"))
    is_md = (
        effective_filename.endswith(".md")
        or os.path.basename(effective_filename) == ".cursorrules"
    )
    is_package_config_yaml = effective_filename.endswith(".nuitka-package.config.yml")

    cleanupWindowsNewlines(filename, effective_filename)
    cleanupTrailingWhitespace(filename)
    cleanupWindowsNewlines(filename, effective_filename)

    if is_rst:
        cleanupRstFmt(
            logger=logger,
            filename=filename,
            effective_filename=effective_filename,
            assume_yes_for_downloads=assume_yes_for_downloads,
        )

    if is_md:
        cleanupMarkdownFmt(
            logger=logger,
            filename=filename,
            assume_yes_for_downloads=assume_yes_for_downloads,
        )

    if is_package_config_yaml:
        formatYaml(
            logger=logger,
            path=filename,
            assume_yes_for_downloads=assume_yes_for_downloads,
            ignore_diff=ignore_yaml_diff,
        )
        cleanupWindowsNewlines(filename, effective_filename)
        cleanupTrailingWhitespace(filename)
        checkYamlSchema(
            logger=logger,
            filename=filename,
            effective_filename=effective_filename,
            update=True,
        )
        formatYaml(
            logger=logger,
            path=filename,
            assume_yes_for_downloads=assume_yes_for_downloads,
            ignore_diff=ignore_yaml_diff,
        )
        cleanupWindowsNewlines(filename, effective_filename)
        cleanupTrailingWhitespace(filename)


def formatImage(filename, logger):
    """Format image files (PNG, JPEG).

    Args:
        filename: path to the file
        logger: logger to use
    """
    if filename.endswith(".png"):
        cleanupPngImage(filename, logger=logger)
    else:
        cleanupJpegImage(filename, logger=logger)


def formatJson(filename, effective_filename, assume_yes_for_downloads):
    """Format JSON files."""
    formatJsonFile(filename, assume_yes_for_downloads=assume_yes_for_downloads)
    cleanupTrailingWhitespace(filename)
    cleanupWindowsNewlines(filename, effective_filename=effective_filename)


def autoFormatFile(
    filename,
    git_stage,
    check_only=False,
    effective_filename=None,
    trace=True,
    limit_yaml=False,
    limit_python=False,
    limit_c=False,
    limit_rst=False,
    limit_md=False,
    limit_json=False,
    ignore_errors=False,
    ignore_yaml_diff=True,
    assume_yes_for_downloads=False,
):
    """Format source code with external tools.

    Args:
        filename: str - filename to work on
        git_stage: bool - indicate if this is to be done on staged content
        check_only: bool - indicate if only checking is to be done
        effective_filename: str - derive type of file from this name
        trace: bool - indicate if progress should be traced
        limit_yaml: bool - limit to YAML files
        limit_python: bool - limit to Python files
        limit_c: bool - limit to C files
        limit_rst: bool - limit to RST files
        limit_md: bool - limit to MD files
        limit_json: bool - limit to JSON files
        ignore_errors: bool - ignore errors during formatting
        ignore_yaml_diff: bool - ignore diffs in YAML files
        assume_yes_for_downloads: bool - assume yes for tool downloads

    Returns:
        bool: True if changes were made (or if check failed), False otherwise.
    """

    # pylint: disable=too-many-arguments,too-many-branches,too-many-locals,too-many-statements

    if effective_filename is None:
        effective_filename = filename

    if os.path.isdir(effective_filename):
        return False

    if git_stage:
        old_code = getFileHashContent(git_stage["dst_hash"])
    else:
        old_code = getFileContents(filename, "rb")

    changed = False

    with withTemporaryFile(
        mode="wb", delete=False, suffix=getFilenameExtension(filename)
    ) as output_file:
        tmp_filename = output_file.name
        output_file.write(old_code)
        output_file.close()

        is_c = effective_filename.endswith((".c", ".h"))
        is_cpp = effective_filename.endswith((".cpp", ".h"))
        is_json = effective_filename.endswith(".json")
        is_png = effective_filename.endswith(".png")
        is_jpeg = effective_filename.endswith((".jpeg", ".jpg"))

        is_python = not (
            is_c or is_cpp or is_json or is_png or is_jpeg
        ) and isPythonFile(filename=tmp_filename, effective_filename=effective_filename)

        if not (is_python or is_c or is_cpp or is_json or is_png or is_jpeg):
            # Text check
            is_txt = effective_filename.endswith(
                (
                    ".patch",
                    ".txt",
                    ".qml",
                    ".rst",
                    ".inc",
                    ".sh",
                    ".in",
                    ".md",
                    ".toml",
                    ".asciidoc",
                    ".nuspec",
                    ".yml",
                    ".stylesheet",
                    ".j2",
                    ".gitignore",
                    ".gitattributes",
                    ".gitmodules",
                    ".spec",
                    "-rpmlintrc",
                    # spell-checker: ignore rpmlintrc
                    "Containerfile",
                    ".containerfile",
                    ".containerfile.in",
                    ".1",
                    ".pth",
                    ".ps1",
                )
            ) or os.path.basename(filename) in (
                "changelog",
                "compat",
                "control",
                "copyright",
                "lintian-overrides",
                ".cursorrules",
            )
        else:
            is_txt = False

        if limit_yaml or limit_python or limit_c or limit_rst or limit_md or limit_json:
            is_match = False

            if effective_filename.endswith(".nuitka-package.config.yml") and limit_yaml:
                is_match = True
            elif (is_c or is_cpp) and limit_c:
                is_match = True
            elif is_python and limit_python:
                is_match = True
            elif effective_filename.endswith((".rst", ".inc")) and limit_rst:
                is_match = True
            elif (
                effective_filename.endswith(".md")
                or os.path.basename(effective_filename) == ".cursorrules"
            ) and limit_md:
                is_match = True
            elif is_json and limit_json:
                is_match = True

            if not is_match:
                is_python = is_c = is_cpp = is_txt = is_json = is_png = is_jpeg = False

        if not (is_python or is_c or is_cpp or is_txt or is_json or is_png or is_jpeg):
            deleteFile(tmp_filename, must_exist=True)
            return False

        if is_python:
            if not _shouldNotFormatCode(
                filename=tmp_filename, effective_filename=effective_filename
            ):
                formatPython(
                    logger=tools_logger,
                    filename=tmp_filename,
                    effective_filename=effective_filename,
                    ignore_errors=ignore_errors,
                    assume_yes_for_downloads=assume_yes_for_downloads,
                )
        elif is_c or is_cpp:
            if not _shouldNotFormatCode(
                filename=tmp_filename, effective_filename=effective_filename
            ):
                formatC(
                    logger=tools_logger,
                    filename=tmp_filename,
                    effective_filename=effective_filename,
                    check_only=check_only,
                    assume_yes_for_downloads=assume_yes_for_downloads,
                    reject_message="Formatting C files needs 'clang-format'.",
                )
        elif is_txt:
            if not _shouldNotFormatCode(
                filename=tmp_filename, effective_filename=effective_filename
            ):
                formatText(
                    logger=tools_logger,
                    filename=tmp_filename,
                    effective_filename=effective_filename,
                    ignore_yaml_diff=ignore_yaml_diff,
                    assume_yes_for_downloads=assume_yes_for_downloads,
                )
        elif is_json:
            formatJson(
                tmp_filename,
                effective_filename=effective_filename,
                assume_yes_for_downloads=assume_yes_for_downloads,
            )
        elif is_png or is_jpeg:
            formatImage(tmp_filename, logger=tools_logger)

        if is_python:
            _transferBOM(filename, tmp_filename)

        changed = old_code != getFileContents(tmp_filename, "rb")

        if changed:
            if check_only:
                my_print("%s: FAIL." % filename, style="red")
            else:
                if trace:
                    my_print("Updated %s." % filename)

                with withPreserveFileMode(filename):
                    if git_stage:
                        new_hash_value = putFileHashContent(tmp_filename)
                        updateFileIndex(git_stage, new_hash_value)
                        updateGitFile(
                            filename, git_stage["dst_hash"], new_hash_value, staged=True
                        )
                    else:
                        shutil.copy(tmp_filename, filename)

    deleteFile(tmp_filename, must_exist=True)

    return changed


@contextlib.contextmanager
def withFileOpenedAndAutoFormatted(
    filename,
    effective_filename,
    ignore_errors=False,
    check_only=False,
    post_format_hook=None,
):
    """Context manager for opening a file and auto-formatting it on close.

    Args:
        filename: path to the file
        effective_filename: str - filename to use for formatting decisions
        ignore_errors: bool - if errors should be ignored
        check_only: bool - if true, check for diff without modifying the file
        post_format_hook: callable - optional hook called with tmp_filename before diff check
    """

    with withTemporaryFilename(
        prefix=os.path.basename(filename),
        suffix=getFilenameExtension(filename) or ".tmp",
    ) as tmp_filename:
        with openTextFile(tmp_filename, "w") as output:
            yield output

        autoFormatFile(
            filename=tmp_filename,
            git_stage=False,
            effective_filename=effective_filename,
            trace=False,
            ignore_errors=ignore_errors,
            assume_yes_for_downloads=True,
        )

        if os.name == "nt":
            autoFormatFile(
                filename=tmp_filename,
                git_stage=False,
                effective_filename=effective_filename,
                trace=False,
                ignore_errors=ignore_errors,
                assume_yes_for_downloads=True,
            )

        if post_format_hook is not None:
            post_format_hook(tmp_filename)

            # The hook might have added something that messes with the formatting, so run it again
            autoFormatFile(
                filename=tmp_filename,
                git_stage=False,
                effective_filename=effective_filename,
                trace=False,
                ignore_errors=ignore_errors,
                assume_yes_for_downloads=True,
            )

        if not os.path.exists(filename) or getFileContents(
            tmp_filename, mode="rb"
        ) != getFileContents(filename, mode="rb"):
            if check_only:
                tools_logger.sysexit(
                    "Error, generated code for %s would change." % effective_filename
                )

            with withPreserveFileMode(filename):
                shutil.copy(tmp_filename, filename)


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
